#!/usr/local/bin/wish8.3


#Copyright (c) 2001 Amar Chaudhary. All rights reserved.
#Copyright (c) 1998-2001 Regents of the University of California.
#All rights reserved.
#
#     ``The contents of this file are subject to the Open Sound World Public
#     License Version 1.0 (the "License"); you may not use this file except in
#     compliance with the License. A copy of the License should be included
#     in a file named "License" or "License.txt" in the distribution from 
#     which you obtained this file. 
#
#     Software distributed under the License is distributed on an "AS IS"
#     basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
#     License for the specific language governing rights and limitations
#     under the License.
#
#     The Original Code is Open Sound World (OSW) software.
#
#     The Initial Developer of the Original Code is Amar Chaudhary.
#     Portions created by Amar Chaudhary are Copyright (C) 1998-2001 
#     Regents of the University of California and Amar Chaudhary. 
#     All Rights Reserved.
#
#     Contributor(s):


if {$tcl_platform(platform) == "windows"} {
    package require registry 1.0
    set env(OSW_DIR) [registry get HKEY_LOCAL_MACHINE\\Software\\OSW Home]
}


lappend auto_path $env(OSW_DIR)
package require Iwidgets
package require xml
catch {namespace import ::itcl::*}


#this one has to come first
source $env(OSW_DIR)/scripts/evalserver.tcl



switch $tcl_platform(platform) windows {
    load $env(OSW_DIR)/libosw.dll Osw
} unix {
    load $env(OSW_DIR)/libosw.so Osw
} default {
    error "Your platform is not supported!"
    exit
}




set oswversion [oswVersion]
source $env(OSW_DIR)/scripts/bugfix.tcl
source $env(OSW_DIR)/scripts/wirecolor.tcl
source $env(OSW_DIR)/scripts/ldelete.tcl
source $env(OSW_DIR)/scripts/commonStr.tcl
source $env(OSW_DIR)/scripts/changeColor.tcl
source $env(OSW_DIR)/scripts/extmanage.tcl
source $env(OSW_DIR)/scripts/balloon.tcl
source $env(OSW_DIR)/scripts/completionbox.tcl
source $env(OSW_DIR)/scripts/font.tcl
    
source $env(OSW_DIR)/scripts/transform.tcl
source $env(OSW_DIR)/scripts/connections.tcl
source $env(OSW_DIR)/scripts/stickynote.tcl
source $env(OSW_DIR)/scripts/patch.tcl
source $env(OSW_DIR)/scripts/oswxml.tcl
source $env(OSW_DIR)/scripts/dock.tcl

set defaultCursor {}
set PasteContextX 0
set PasteContextY 0
set PasteContextSeedX 0
set PasteContextSeedY 0


set oswGravityWell 10


###################



set Selection {}
set oswClipboard {}


global OldX OldY
proc StartObjectMotion {patch win tag x y shift} {
    
    
    global Selection OldX OldY
    
    
    set OldX [$win canvasx $x]
    set OldY [$win canvasy $y]
    #set OldX $x
    #set OldY $y
    $patch ObjectSelection $tag $shift
    $win bind $tag <B1-Motion> [list ::$tag MoveObject %x %y]
    $win bind $tag <ButtonRelease-1> [list StopObjectMotion $win $tag $shift]
}



proc MoveOthers {moved dx dy} {
    global Selection 
    
    foreach sel $Selection {
	if {"::$sel" != $moved} {
	    switch [$sel info class] "::TransformWidget" {
		$sel MoveObjectRel $dx $dy
	    } "::stickynote" {
		$sel MoveObjectRel $dx $dy              
	    } "::Connection" {
		#$sel Move $dx $dy
	    } default {
		$sel MoveObjectRel $dx $dy
	    }
	}
    }
}


proc StopObjectMotion {win tag shift} {
    global Selection
    $win bind $tag <B1-Motion> {}
    $win bind $tag <ButtonRelease-1> {}
    if {[llength $Selection] == 1 && [$tag info class] == "::TransformWidget"} {
	$Patch::selected SplitWire $tag 
    }
}



proc SetEditMenu {onoff} {
    global Selection oswClipboard
    if {$onoff && [llength $Selection] > 0} {
	.patchmenu.edit entryconfigure "Cut" -state normal
	.patchmenu.edit entryconfigure "Copy" -state normal
	.patchmenu.edit entryconfigure "Clear" -state normal
	#.patchmenu.edit entryconfigure "Collapse Subpatch" -state normal
	#.patchmenu.format configure -state normal
    } else {
	.patchmenu.edit entryconfigure "Cut" -state disabled
	.patchmenu.edit entryconfigure "Copy" -state disabled
	.patchmenu.edit entryconfigure "Clear" -state disabled
	#.patchmenu.edit entryconfigure "Collapse Subpatch" -state disabled
	#.patchmenu.format configure -state disabled
    }
    if {[llength $oswClipboard] > 0} {
	.patchmenu.edit entryconfigure "Paste" -state normal
    } else {
	.patchmenu.edit entryconfigure "Paste" -state disabled
    }
}


proc Help {} {
    global Selection
    foreach sel $Selection {
	#Try getting help for the first selected transform
	if {[$sel info class] == "::TransformWidget"} {
	    $sel HelpPatch
	    $sel Help
	    return
	}
    }
    
    .help activate
}


proc HelpPatch {} {
    global Selection
    foreach sel $Selection {
	#Try getting help for the first selected transform
	if {[$sel info class] == "::TransformWidget"} {
	    $sel HelpPatch
	    return
	}
    }
}


proc ShowDock {} {
    global showdock
    if {$showdock} {
	wm deiconify .dock
    } else {
	wm withdraw .dock
    }


}


proc GetTransformProperties {} {
    global Selection
    foreach sel $Selection {
	#Try getting help for the first selected transform
	if {[$sel info class] == "::TransformWidget"} {
	    $sel OpenDialog
	    return
	}
    }
    
    .help activate
}


proc StartExternalizerFromTransform {} {
    global Selection initCmd
    foreach sel $Selection {
	if {[$sel info class] == "::TransformWidget"} {
	    StartExternalizer [oswFindTransformSpecification \
		    [lindex $initCmd(::$sel) 0]]
	    return
	}
    }


    error "No transform selected"
}


proc StartExternalizer {args} {
    global tcl_platform env


    if {[llength $args] > 0} {
	switch $tcl_platform(platform) windows {
	    exec wish82 $env(OSW_DIR)/externalizer/externalizer.tcl \
		    [lindex $args 0] -calledFromOsw &
	} unix {
	    exec wish8.3 $env(OSW_DIR)/externalizer/externalizer.tcl \
		    [lindex $args 0] -calledFromOsw &
	}
    } else {
	switch $tcl_platform(platform) windows {
	    exec wish82 $env(OSW_DIR)/externalizer/externalizer.tcl &
	} unix {
	    exec wish8.3 $env(OSW_DIR)/externalizer/externalizer.tcl &
	} 
    }
}


##################

#this functions were originally for a project in an HCI class that involved
#recording and analyzing user input.  They might come in handy again someday.

set isRecording false
set logfile {}
proc RecordInput {} {



    if {$isRecording} {
	StopRecording
    }     


    set filename [tk_getSaveFile -defaultextension "txt" -filetypes {{"Log File" {.txt}}} ]
    
    if {$filename != ""} {
	set logfile [open $filename w]
	bind all <Button> {catch {puts $logfile [list %t %T %W %b %x %y %s]}}
	bind all <Motion> {catch {puts $logfile [list %t %T %W %b %x %y %s]}}
	bind all <ButtonRelease> {catch {puts $logfile [list %t %T %W %b %x %y %s]}}
	bind all <KeyPress> {catch {puts $logfile [list %t %T %W %k %s]}}
	bind all <KeyRelease> {catch {puts $logfile [list %t %T %W %k %s]}}
	
	set isRecording true
    }
}


proc StopRecording {} {


    set isRecording false
    bind all <Button> {}
    bind all <Motion> {}
    bind all <ButtonRelease> {}
    bind all <KeyPress> {}
    bind all <KeyRelease> {}
    if {$logfile != {}} {
	close $logfile
    }
}


proc PlayInput {} {
    set filename [tk_getOpenFile -defaultextension "txt" -filetypes {{"Log File" {.txt}}} ]
    
    if {$filename != ""} {
	set playfile [open $filename r]
	while {![eof $playfile]} {
	    after 20
	    set line [gets $playfile]
	    switch {[lindex $line 1]} 2 {
		event generate [lindex $line 2] <KeyPress> \
			-keycode [lindex $line 3] -state [lindex $line 4]
	    } 3 {
		event generate [lindex $line 2] <KeyRelease> \
			-keycode [lindex $line 3] -state [lindex $line 4]
	    } 4 {
		event generate [lindex $line 2] <Button> \
			-button [lindex $line 3] -x [lindex $line 4] \
			-y [lindex $line 5] -state [lindex $line 6]
	    } 5 {
		event generate [lindex $line 2] <ButtonRelease> \
			-button [lindex $line 3] -x [lindex $line 4] \
			-y [lindex $line 5] -state [lindex $line 6]
	    } 6 {
		event generate [lindex $line 2] <Motion> \
			-button [lindex $line 3] -x [lindex $line 4] \
			-y [lindex $line 5] -state [lindex $line 6]
	    } 
	}
    }
}


###########################


proc oswSetFromGlobalVariable {toSet varName} {
    
    global $varName


    oswSet $toSet [set $varName]
}


set DeferredEvalVar ""
trace variable DeferredEvalVar w oswDoDeferred


proc oswDoDeferred {name1 name2 op} {
    eval $::DeferredEvalVar
}

proc oswSetWidgetProperty {widget property args} {
    $widget configure $property $args
}

proc oswSetFromTextBox {toSet textbox} {
    oswSet $toSet [$textbox get]
}


proc oswFixPatchTerminals {patch} {
    patchObj-$patch FixPatchTerminals
}


proc oswFixTransformTerminals {transform} {
    set patch [file dir $transform]
    set transformObj [patchObj-${patch} Map $transform]
    $transformObj AppendTerminals
}


####################


set showdock 1

proc oswAttachDefaultBindings {w} {
    bind $w <Control-n> {oswNewPatch}
    bind $w <Control-q> {FileQuit}
    bind $w <Control-s> {if {$Patch::selected != ""} {$Patch::selected Save}}
    bind $w <Control-o> {if {$Patch::selected != ""} {oswOpenPatch $Patch::selected}}
    bind $w <Control-w> {if {$Patch::selected != ""} {oswClosePatch $Patch::selected}}
    bind $w <Delete> {if {$Patch::selected != ""} {$Patch::selected DeleteSelection}}
    bind $w <Control-x> {if {$Patch::selected != ""} {$Patch::selected Cut}}
    bind $w <Control-c> {if {$Patch::selected != ""} {$Patch::selected Copy}}
    bind $w <Control-v> {if {$Patch::selected != ""} {$Patch::selected Paste}}
    bind $w <Control-a> {if {$Patch::selected != ""} {$Patch::selected SelectAll}}
    bind $w <Control-m> {if {$Patch::selected != ""} {$Patch::selected AddNewTransformFromKeyboard}}
    bind $w <Alt-Return> {GetTransformProperties}
    bind $w <Control-d> {if {$Patch::selected != ""} {$Patch::selected DockSelection}}
    bind $w <F5> {StartExternalizerFromTransform}
    bind $w <F1> Help
    bind $w <Shift-F1> HelpPatch
    bind $w <Control-Prior> {if {$Patch::selected != ""} {$Patch::selected ScaleUp}}
    bind $w <Control-Next> {if {$Patch::selected != ""} {$Patch::selected ScaleDown}}
}


proc oswInitWindowControl {} {
    bind . <Map> oswShowAllRelevant
    bind .dock <Map> oswShowAllRelevant
}

set _inShowAllRelevant 0
proc oswShowAllRelevant {} {
    if {!$::_inShowAllRelevant} {
	set ::_inShowAllRelevant 1
	wm deiconify .
	if {$Patch::selected != ""} {
	    raise [$Patch::selected cget -pane]
	}
	raise .
	ShowDock
	if {$::showdock} {
	    raise .dock
	}
	set ::_inShowAllRelevant 0
    }
}

proc oswInitAppFrame {} {


    label .status -text ""
    label .status2 -text ""
    #pack .status2 .status -side bottom -fill x


    menu .patchmenu -tearoff 0
    .patchmenu add cascade -label "File" -menu .patchmenu.file
    menu .patchmenu.file -tearoff 0
    .patchmenu.file add command -label "New" -command {oswNewPatch} \
	    -accelerator "Ctrl N"
    .patchmenu.file add command -label "Open" -command {oswOpenPatch $Patch::selected} \
	    -accelerator "Ctrl O"
    .patchmenu.file add command -label "Close" -command {oswClosePatch $Patch::selected} \
	    -accelerator "Ctrl W"
    .patchmenu.file add command -label "Save" -command {$Patch::selected Save} \
	    -accelerator "Ctrl S"
    .patchmenu.file add command -label "Save As..." -command {$Patch::selected SaveAs}
    .patchmenu.file add separator
    
    .patchmenu.file add command -label "Quit" -command "FileQuit" \
	    -accelerator "Ctrl Q"
    . configure -menu .patchmenu
        
    .patchmenu add cascade -label "Edit" -menu .patchmenu.edit
    menu .patchmenu.edit -tearoff 0
    .patchmenu.edit add command -label "Select All" -command {$Patch::selected SelectAll} \
	    -accelerator "Ctrl A" 
    .patchmenu.edit add command -label "Cut" -command {$Patch::selected Cut} \
	    -accelerator "Ctrl X" -state disabled
    .patchmenu.edit add command -label "Copy" -command {$Patch::selected Copy} \
	    -accelerator "Ctrl C" -state disabled
    .patchmenu.edit add command -label "Paste" -command {$Patch::selected Paste} \
	    -accelerator "Ctrl V" -state disabled
    .patchmenu.edit add command -label "Clear" -accelerator "Del" \
	    -command {$Patch::selected DeleteSelection} -state disabled
    .patchmenu.edit add separator
    .patchmenu.edit add command -label "Insert Blank Transform " \
	    -command {$Patch::selected AddNewTransformFromKeyboard} \
	    -accelerator "Ctrl M"
    .patchmenu.edit add command -label "Insert Annotation" \
	    -command {$Patch::selected DragNewSticky} \


    #.patchmenu.edit add separator
    #.patchmenu.edit add command -label "Collapse Subpatch" \
    #	    -command {$Patch::selected Collapse} \
    #	    -accelerator "Ctrl K" 
    
    
    
    .patchmenu add cascade -label "View" -menu .patchmenu.view
    menu .patchmenu.view -tearoff 0
    #.patchmenu.view add command -label "Zoom In" -command {$Patch::selected ScaleUp}
    #.patchmenu.view add command -label "Zoom Out" -command {$Patch::selected ScaleDown}
    #.patchmenu.view add command -label "Split" \
    #	    -command ShowAHiddenView
    #.patchmenu.view add separator
    .patchmenu.view add checkbutton -label "Show Dock" -variable showdock -command "ShowDock"
    
    #bind . <Control-Prior> {$Patch::selected ScaleUp}
    #bind . <Control-Next> {$Patch::selected ScaleDown}


    .patchmenu add cascade -label "Format" -menu .patchmenu.format
    menu .patchmenu.format -tearoff 0
    
    .patchmenu.format add command -label "Color..." \
	    -command {$Patch::selected SelectTransformColor}
    .patchmenu.format add cascade -label "Font Face" -menu .patchmenu.format.font
    .patchmenu.format add cascade -label "Font Size" -menu .patchmenu.format.size
    menu .patchmenu.format.font -tearoff 0
    foreach face [lsort [font families]] {
	.patchmenu.format.font add radiobutton -label $face -command {$Patch::selected SelectTransformFontFace $_oswFontFace} -variable _oswFontFace
    }
    
    menu .patchmenu.format.size -tearoff 0
    foreach size {8 10 12 14 16 18 20 24 28 32 26 40 48 56 64 72} {
	.patchmenu.format.size add radiobutton -label $size -command {$Patch::selected SelectTransformFontSize $_oswFontSize} -variable _oswFontSize
    }


    #.patchmenu add cascade -label "Transform" -menu .patchmenu.transform
    menu .patchmenu.transform -tearoff 0
    .patchmenu.transform add command -label "Properties..." \
	    -accelerator "Alt Enter" \
	    -command "GetTransformProperties"
    .patchmenu.transform add command -label "Dock" \
	    -accelerator "Ctrl D" \
	    -command {$Patch::selected DockSelection}
    .patchmenu.transform add command -label "View Specification..." \
	    -command {StartExternalizerFromTransform} -accelerator "F5"
        
    .patchmenu add cascade -label "Patch" -menu .patchmenu.patch
    menu .patchmenu.patch -tearoff 0
    #.patchmenu.patch add command -label "Arguments..." \
	    #       -command CreateArguments]
    .patchmenu.patch add command -label "Background..." \
	    -command {$Patch::selected SelectBackground}


    .patchmenu add cascade -label "Options" -menu .patchmenu.options
    menu .patchmenu.options -tearoff 0
    .patchmenu.options add command -label "Audio Devices..." -command oswAudioPrefs
    .patchmenu.options add command -label "Editor..." -command oswEditorPrefs
    .patchmenu.options add separator
    .patchmenu.options add checkbutton -label "Profiling" -variable oswProfiling -command oswToggleProfiling -accelerator "Ctrl Shift P"
    .patchmenu.options add command -label "Dump Profile" -command oswDumpProfileToFile
    .patchmenu.options add separator
    .patchmenu.options add command -label "Reset Main Clock" -command oswResetMainClock
    #bind . <Control-P> {set ::oswProfiling [expr !$::oswProfiling] ; oswToggleProfiling}


    .patchmenu add cascade -label "Help" -menu .patchmenu.help
    menu .patchmenu.help -tearoff 0
    .patchmenu.help add command -label "Help Topic" -accelerator "F1" \
	    -command Help
    #.patchmenu.help add command -label "Help Patch" -accelerator "Shift F1" \
#	    -command HelpPatch
    .patchmenu.help add separator
    .patchmenu.help add command -label "User Guide" -command {ShowHelpTopic UserGuide}
    .patchmenu.help add command -label "Transforms (by function)" -command {ShowHelpTopic Transform_by_group}
    .patchmenu.help add command -label "Transforms (alphabetical)" -command {ShowHelpTopic Transform_by_alpha}
    .patchmenu.help add command -label "Browse Help Patches..." -command {BrowseHelpPatches $Patch::selected}
    .patchmenu.help add command -label "Browse Tutorials..." -command {BrowseTutorials $Patch::selected}
    .patchmenu.help add separator
    .patchmenu.help add checkbutton -label "Show Balloon Help" -variable oswShowBalloon
    .patchmenu.help add checkbutton -label "Show Command Completions" -variable oswShowCompletions
    .patchmenu.help add separator
    .patchmenu.help add command -label "About OSW..." -command {AboutBox 0}
    
    oswAttachDefaultBindings .
    bind . <Button> oswShowAllRelevant
    wm geometry . [expr [winfo screenwidth .] - 10]x0+0+0
    #bind . <Destroy> {if {!$_quitting} {oswQuit}}
    #iwidgets::panedwindow .pw  -orient vertical -width 800 -height 600
    #.pw add "dock"
    #.pw add "patches"


    #set ::patchArea [iwidgets::panedwindow [.pw childsite patches].pw \
	    #-orient horizontal]
    #pack $::patchArea -side top -fill both -expand yes
    
    #pack .pw -side top -fill both -expand yes


    #if {$::showdock} {
#	wm deiconify .dock
#    } else {
#	wm withdraw .dock
#    }
}


################### 


set oswShowBalloon 0
set oswShowCompletions 1

balloon TheBalloonWindow

proc oswShowBalloonHelp {text x y W} {
    if {$::oswShowBalloon} {
	TheBalloonWindow Show $x $y $text $W
    }
}

proc oswHideBalloonHelp {} {
   if {$::oswShowBalloon} {
	TheBalloonWindow Hide
    }
}


################### 


set defaultCount -1
proc oswNewPatch {} {
    global defaultCount


    incr defaultCount
    Patch patchObj-/default$defaultCount -patchClass default -patchInstance default$defaultCount 
    patchObj-/default$defaultCount Unlock
    #wm deiconify .default$defaultCount
    #focus .default$defaultCount


}



proc oswOpenPatch {patch} {
    global defaultCount

    set patchFile [tk_getOpenFile -defaultextension "osw" -filetypes {{"OSW Patch" {.osw}}} ]


    if {$patchFile == ""} {
	return
    }
    
    #$patch Hide
    oswReadPatchFromXML $patchFile [list "" "" "" $patchFile]
}


proc oswReadPatchFromXML {patchFile instanceInfo} {
    # Get the document data, ensuring that UTF-8 characters are read
    
    set ch [open $patchFile]
    fconfigure $ch -encoding utf-8
    set data [read $ch]
    close $ch


    oswParseXML $data $instanceInfo
}


proc oswParseXML {data instanceInfo} {
    if {[llength $instanceInfo] > 0} {
	PushXMLState PATCHINSTANCE $instanceInfo
    }


    if {[catch {$::xmlparser parse $data} err]} {
	tk_messageBox -parent . -type ok -title "Error" \
		-message "Unable to display patch due to \"$err\"" \
		-icon error
    }


    if {[llength $instanceInfo] > 0} {
	PopXMLState
    }
    
}


proc oswClosePatch {patch} {
    if {[$patch cget -parent] == ""} {
	delete object $patch
	set Patch::selected $Patch::previousSelected
    } else {
	wm withdraw [$patch cget -win]
    }
}


set NumPatches 0


proc oswInstancePatch {parent patchClass patchInstance patchCmd patchPath args} {
    global env

    #puts "$patchClass $patchInstance \"$patchCmd\" $patchPath"
    set patchArgs [lrange $patchCmd 2 end]
    if {[file exists $patchClass.osw]} {
	#puts "QQ $patchClass $patchInstance"
	oswReadPatchFromXML $patchClass.osw [list $parent $patchInstance \
		$patchArgs ""]
	return true
    }
    if {[file exists $patchPath]} {
	#puts "RR $patchClass $patchInstance"
	oswReadPatchFromXML $patchPath [list $parent $patchInstance \
		$patchArgs ""]
	return true
    }
    if {[file exists $env(OSW_DIR)/patches/$patchClass.osw]} {
	oswReadPatchFromXML $osw(OSW_DIR)/patches/$patchClass.osw \
		[list $parent $patchInstance $patchArgs ""]
	return true
    }


    return false
}


proc oswAddSubpatch {parent class} {
    oswParseXML "<PATCH CLASS=\"$class\"></PATCH>" [list $parent "" ""]
    patchObj-$parent/$class Unlock
}


proc oswSearchExternalPathAux {path fname} {
    if {[file exists $path/$fname]} {
	return $path/$fname
    } else {
	foreach maybeDir [glob -nocomplain $path/*] {
	    if {[file isdirectory $maybeDir]} {
		set result [oswSearchExternalPathAux $maybeDir $fname]
		if {$result != ""} {
		    return $result
		}
	    }
	}
	return ""
    }
}


proc oswSearchExternalPath {fname} {
    global env
    set found [oswSearchExternalPathAux $env(OSW_DIR)/externals $fname]
    if {$found == ""} {
	set found [oswSearchExternalPathAux $env(OSW_DIR)/tutorials $fname]
    }           
    return $found
}


proc oswLinkTo {link} {
    set patchLink $link
    append patchLink ".osw"
    set found [oswSearchExternalPath $patchLink]
    if {$found != ""} {
	oswSourcePatch $found
	return
    }       
    ShowHyperlink $link
}



proc oswLoadBitmap {type fname args} {


    set name [file root [file tail $fname]]
    if {[llength $args] > 0} {
	global oswBitmap_$name
    }
    global env
        
    if {[file exists $fname]} {
	set oswBitmap_$name [image create $type -file $fname]
    } elseif {[file exists $env(OSW_DIR)/bitmaps/$fname]} {
	set oswBitmap_$name [image create $type -file $env(OSW_DIR)/bitmaps/$fname]
    } else {
	if {[catch {source [oswSearchExternalPath $scriptFile]}]} {
	    bgerror "Couldn't find script file $fname"
	    return
	}
    }
    return [set oswBitmap_$name]
}


proc oswLoadBitmapFromData {name type data} {
    
    
    global oswBitmap_$name env
    
    
    if {[info exists oswBitmap_$name]} {
	return [set oswBitmap_$name]
    }
    
    
    set oswBitmap_$name [image create $type -data $data]
    
    
    return set [set oswBitmap_$name]
}


proc oswSourcePatch {patchFile} {
    
    set patchName [file rootname [file tail $patchFile]]
    set instanceName $patchName
    #if {[info exists .$patchName]} {
	# there is already a toplevel patch $patchName
	#    append instanceName $defaultCount
	#    incr defaultCount
	#   }
    oswReadPatchFromXML $patchFile [list "" $instanceName {} $patchFile]
}
    
    
proc oswSourcePatchTopLevel {patchFile} {
    regsub "^OSW_DIR" $patchFile $::env(OSW_DIR) patchFile
    oswSourcePatch $patchFile
}


proc oswShowPatch {patchObj} {
    #wm deiconify [$window cget -win]
    $patchObj Show
}



set _scripts {}
proc oswLoadScript {scriptFile} {
    
    
    global _scripts env
    
    
    if {[lsearch $_scripts $scriptFile] != -1} {
	return
    }
    
    
    if {[file exists $scriptFile]} {
	source $scriptFile
    } elseif {[file exists $env(OSW_DIR)/scripts/$scriptFile]} {
	source $env(OSW_DIR)/scripts/$scriptFile
    } else {
	if {[catch {source [oswSearchExternalPath $scriptFile]}]} {
	    bgerror "Couldn't find script file $fname"
	    return
	}
    }
    
    
    lappend _scripts $scriptFile
}    

set _quitting 0
proc FileQuit {} {
    global _quitting
    if {!$_quitting} {
	set _quitting 1
	#oswDestroyAllPatches
	oswQuit
	exit
	#destroy .
    }
}


################


set oswProfiling 0


proc oswToggleProfiling {} {
    if {$::oswProfiling} {
	puts "Profiling On"
    } else {
	puts "Profiling Off"
    }
    oswProfile $::oswProfiling
}


proc oswDumpProfileToFile {} {
    set filename [tk_getSaveFile -defaultextension "txt" -filetypes {{"Text File" {.txt}}} ]
    if {$filename != ""} {
	puts "Dumping Profile to $filename"
	oswDumpProfile $filename
    }
}


proc oswResetMainClock {} {
    oswSet /main_clock 0.0
}


################


source $env(OSW_DIR)/scripts/audioprefs.tcl
source $env(OSW_DIR)/scripts/editorprefs.tcl


oswLoadScript splash.tcl
wm withdraw .
wm title . "Open Sound World"
AboutBox 1
after 3000 destroy .aboutosw
oswInitAppFrame
wm deiconify .

source $env(OSW_DIR)/scripts/help.tcl


oswInitDock

if {[llength $argv] > 0} {
    foreach patch $argv {
	if {[file extension $patch] == ".osx"} {
	    oswLoadOSX $patch
	    oswNewPatch
	} elseif {[file extension $patch] == ".osd"} {
	    StartExternalizer $patch
	} else {
	    oswSourcePatch $patch
	}
    }
} else {
    oswNewPatch
}

oswInitWindowControl

oswMainLoop

